import cython
import numpy as np
cimport numpy as np
cimport ccoldatoms_lib
import atexit
from libc.stdint cimport uintptr_t


@cython.boundscheck(False)
@cython.wraparound(False)
def coulomb_force(
    np.ndarray[double, ndim=2, mode="c"] positions not None,
    double charge, double dt, double delta, double k,
    np.ndarray[double, ndim=2, mode="c"] forces not None):

    cdef num_ptcls
    num_ptcls = positions.shape[0]

    ccoldatoms_lib.ca_coulomb_force(
        &positions[0, 0], charge, dt, num_ptcls, delta, k, &forces[0, 0])


@cython.boundscheck(False)
@cython.wraparound(False)
def coulomb_force_per_particle_charge(
    np.ndarray[double, ndim=2, mode="c"] positions not None,
    np.ndarray[double, ndim=1, mode="c"] charges not None,
    double dt, double delta, double k,
    np.ndarray[double, ndim=2, mode="c"] forces not None):

    cdef num_ptcls
    num_ptcls = positions.shape[0]

    ccoldatoms_lib.ca_coulomb_force_per_particle_charge(
        &positions[0, 0], &charges[0], dt, num_ptcls, delta, k, &forces[0, 0])


@cython.boundscheck(False)
@cython.wraparound(False)
def harmonic_trap_forces(
    np.ndarray[double, ndim=2, mode="c"] positions not None,
    double q,
    double kx, double ky, double kz, double phi,
    double dt, 
    np.ndarray[double, ndim=2, mode="c"] forces not None):
    assert(positions.shape[0] <= forces.shape[0])
    assert(positions.shape[1] == 3)
    assert(forces.shape[1] == 3)

    cdef num_ptcls
    num_ptcls = positions.shape[0]

    ccoldatoms_lib.ca_harmonic_trap_forces(
        &positions[0, 0], q, kx, ky, kz, phi, dt, num_ptcls, &forces[0, 0])


@cython.boundscheck(False)
@cython.wraparound(False)
def harmonic_trap_forces_per_particle_charge(
    np.ndarray[double, ndim=2, mode="c"] positions not None,
    np.ndarray[double, ndim=1, mode="c"] q not None,
    double kx, double ky, double kz, double phi,
    double dt, 
    np.ndarray[double, ndim=2, mode="c"] forces not None):
    assert(positions.shape[0] <= forces.shape[0])
    assert(positions.shape[1] == 3)
    assert(forces.shape[1] == 3)

    cdef num_ptcls
    num_ptcls = positions.shape[0]

    ccoldatoms_lib.ca_harmonic_trap_forces_per_particle_charge(
        &positions[0, 0], &q[0], kx, ky, kz, phi, dt, num_ptcls, &forces[0, 0])


@cython.boundscheck(False)
@cython.wraparound(False)
def bend_kick_update_scalar(
    double dt, double omegaB,
    np.ndarray[double, ndim=2, mode="c"] x not None,
    np.ndarray[double, ndim=2, mode="c"] v not None):

    cdef num_ptcls
    num_ptcls = x.shape[0]

    ccoldatoms_lib.ca_bend_kick_update_scalar(
        dt, omegaB, num_ptcls, &x[0, 0], &v[0, 0])


@cython.boundscheck(False)
@cython.wraparound(False)
def bend_kick_update_vector(
    double dt,
    np.ndarray[double, ndim=1, mode="c"] omegaB,
    np.ndarray[double, ndim=2, mode="c"] x not None,
    np.ndarray[double, ndim=2, mode="c"] v not None):

    cdef num_ptcls
    num_ptcls = x.shape[0]

    ccoldatoms_lib.ca_bend_kick_update_vector(
        dt, &omegaB[0], num_ptcls, &x[0, 0], &v[0, 0])


def bend_kick_update(dt, omegaB, x, v):
    if np.isscalar(omegaB):
        bend_kick_update_scalar(dt, omegaB, x, v)
    else:
        bend_kick_update_vector(dt, omegaB, x, v)


cdef class Rng(object):

    cdef ccoldatoms_lib.CARandCtx* _generator

    def context(self):
        return <uintptr_t>self._generator

    def __cinit__(self, seed=None):
        self._generator = ccoldatoms_lib.ca_rand_create()
        if seed is not None:
            ccoldatoms_lib.ca_rand_seed(self._generator, seed)

    def __dealloc__(self):
        ccoldatoms_lib.ca_rand_destroy(&self._generator)

    def seed(self, unsigned int seed):
        ccoldatoms_lib.ca_rand_seed(self._generator, seed)

    def fill(self, np.ndarray[double, mode="c"] array):
        cdef int n = array.size
        cdef double *a = <double *>array.data
        ccoldatoms_lib.ca_rand(self._generator, n, a)

    def fill_gaussian(self, double mean, double std, np.ndarray[double, mode="c"] array):
        cdef int n = array.size
        cdef double *a = <double *>array.data
        ccoldatoms_lib.ca_rand_gaussian(self._generator, n, mean, std, a)

    def fill_poisson(self, double nbar, np.ndarray[int, mode="c"] array):
        cdef int n = array.size
        cdef int *a = <int *>array.data
        ccoldatoms_lib.ca_rand_poisson(self._generator, n, nbar, a)


"""For convenience we provide a random number generator."""
rng = Rng()


@cython.boundscheck(False)
@cython.wraparound(False)
def compute_nbars(double dt, double gamma, 
    np.ndarray[double, ndim=1, mode="c"] s_of_r not None,
    np.ndarray[double, ndim=1, mode="c"] delta not None,
    np.ndarray[double, ndim=1, mode="c"] nbar not None):
    assert(s_of_r.shape[0] == nbar.shape[0])
    assert(delta.shape[0] == nbar.shape[0])

    cdef num_ptcls
    num_ptcls = nbar.shape[0]

    ccoldatoms_lib.ca_compute_nbars(num_ptcls, dt, gamma, &s_of_r[0], &delta[0], &nbar[0])

def add_radiation_pressure(
    uintptr_t rng_context,
    np.ndarray[double, ndim=1, mode="c"] hbar_k,
    np.ndarray[double, ndim=1, mode="c"] nbars not None,
    np.ndarray[double, ndim=2, mode="c"] f not None):
    assert(nbars.size == f.shape[0])
    assert(f.shape[1] == 3)
    assert(hbar_k.size == 3)

    cdef num_ptcls
    num_ptcls = nbars.shape[0]
    ccoldatoms_lib.ca_add_radiation_pressure(
        num_ptcls,
        <ccoldatoms_lib.CARandCtx*>rng_context,
        &hbar_k[0],
        &nbars[0],
        &f[0, 0])




cdef extern from "forces.h":
    void ca_coulomb_force(const double* positions, double charge, double dt,
                       int num_ptcls, double delta, double k, double* forces);
    void ca_coulomb_force_per_particle_charge(const double* positions,
                       double* charge, double dt, int num_ptcls, double delta,
                       double k, double* forces);
    void ca_harmonic_trap_forces(const double* positions, double q,
                       double kx, double ky, double kz, double phi, double dt,
                       int num_ptcls, double *forces);
    void ca_harmonic_trap_forces_per_particle_charge(const double* positions, const double* q,
                       double kx, double ky, double kz, double phi, double dt,
                       int num_ptcls, double *forces);


cdef extern from "bend_kick_updater.h":
    void ca_bend_kick_update_scalar(double dt, double omegaB,
        int num_ptcls,
        double *x, double *v);
    
    void ca_bend_kick_update_vector(double dt, const double *omegaB,
        int num_ptcls,
        double *x, double *v);


cdef extern from "ca_rand.h":
    struct CARandCtx:
        pass
    CARandCtx* ca_rand_create();
    void ca_rand_destroy(CARandCtx** ctx);
    void ca_rand_seed(CARandCtx* ctx, int seed);
    void ca_rand(CARandCtx* ctx, int n, double* x);
    void ca_rand_gaussian(CARandCtx* ctx, int n, double mean, double std, double* x);
    void ca_rand_poisson(CARandCtx* ctx, int n, double nbar, int* x);
    void ca_rand_poisson_many(CARandCtx* ctx, int n, double* nbars, int* x);


cdef extern from "radiation_pressure.h":
    void ca_compute_nbars(int n, double dt, double gamma, const double* s_of_r,
        const double* delta, double* nbar);
    void ca_add_radiation_pressure(int n, CARandCtx* ctx, const double* hbar_k, const double* nbar, double* force);



from setuptools import setup
import sys
import os


exec(open('src/coldatoms/version.py').read())


def require_build(argv):
    """This tells us whether we need to build the package."""

    info_commands = ['--help-commands', '--name', '--version', '-V',
                        '--fullname', '--author', '--author-email',
                        '--maintainer', '--maintainer-email', '--contact',
                        '--contact-email', '--url', '--license', '--description',
                        '--long-description', '--platforms', '--classifiers',
                        '--keywords', '--provides', '--requires', '--obsoletes']
    # Add commands that do more than print info, but also don't need Cython and
    # template parsing.
    info_commands.extend(['egg_info', 'install_egg_info', 'rotate'])

    for command in info_commands:
        if command in argv[1:]:
            return False

    return True


def get_package_info():
    """Assembles info for package and external modules."""

    from Cython.Build import cythonize
    from setuptools import find_packages, Extension
    import numpy

    extra_compile_args = [
        '-DDSFMT_MEXP=19937'
        ]
    if sys.platform == 'win32':
        if os.environ['PYTHON_ARCH'] == '64':
            extra_compile_args += [
                '-DHAVE_SSE2'
                ]
    else:
        extra_compile_args += [
            '-DHAVE_SSE2',
            '-std=c99',
            '-ffast-math',
            '-ftree-vectorize',
            '-msse4']

    extra_link_args = []

    coldatoms_lib = cythonize([Extension(
        'coldatoms_lib.coldatoms_lib',
        sources=['src/coldatoms_lib/bend_kick_updater.c',
                'src/coldatoms_lib/forces.c',
                'src/coldatoms_lib/ca_rand.c',
                'src/coldatoms_lib/radiation_pressure.c',
                'src/coldatoms_lib/dSFMT/dSFMT.c',
                'src/coldatoms_lib/coldatoms_lib.pyx'],
        include_dirs=['src/coldatoms_lib/', numpy.get_include()],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args
        )])

    packages = find_packages(where='src',
                             exclude=('tests', 'docs', 'examples'))
    return dict(
        packages=packages,
        package_dir={'': 'src'},
        ext_modules=coldatoms_lib,
    )


def setup_package():
    long_description = """The cold-atoms library is a tool box for the
simulation of ensembles of neutral atoms or ions for atomic, molecular, and
optical physics (AMO) experiments."""
    metadata = dict(
        name='coldatoms',
        maintainer='Dominic Meiser',
        version=__version__,
        description='Collection of python tools for cold atoms simulations',
        long_description=long_description,
        author='Dominic Meiser',
        author_email='dmeiser79@gmail.com',
        url='https://github.com/d-meiser/cold-atoms',
        license='GPLv3',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Education',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Topic :: Scientific/Engineering :: Physics',
        ],
        setup_requires=[
            'cython',
            'numpy',
        ],
        install_requires=[
            'cython',
            'numpy',
        ],

    )
    if require_build(sys.argv):
        metadata.update(get_package_info())
    setup(**metadata)


if __name__ == '__main__':
    setup_package()
